import os
import numpy as np
import plotly.graph_objects as go
import plotly.io as pio
import pandas as pd
# TODO for my testing, not in final 
pio.renderers.default = 'browser'
import json


class evaluate():
    """
    Plot diagnostics and inform user of potential problems in collected
    data.
    
    Parameters
    ----------
    csv_dir : string
        path to diagnostics csv
    
    Attributes
    ----------   
    fs : int 
        sampling rate of rx recordings       
    rx_name : list
         names of rx recordings 
    a_weight : array
         A_Weight of every trial    
    fsf_all : list
         FSF scores of every trial
    peak_dbfs : list
         peak amplitude of each trial, dB relative to full 
         scale      
    trials : int
         number of trials
    clip_flag : list
        Trials that clipped  
    fsf_flag : list
        Trials that have low FSF scores or otherwise deviate 
        from the patterns of the dataset
    aw_flag : list
        Trials that have low dBA values (and likely lost audio)
        or otherwise deviate from the patterns of the dataset     
         
    Methods
    ----------

    See Also
    --------

    Examples
    --------
    
    Returns
    -------
    
    """
    def __init__(self, 
                 csv_dir = '',
                 test_names=None,
                 json_data=None):
        if json_data is None:
            # If only one test, make a list for iterating
            if isinstance(test_names, str):
                test_names = [test_names]
            # Initialize full paths attribute
            self.full_paths = []
            self.test_names = []
            for test_name in test_names:
                # If no extension given use csv
                fname, fext = os.path.splitext(test_name)
                if fext == '':
                    tname = fname + '.csv'
                else:
                    tname = fname + fext
                fpath = os.path.join(csv_dir, 'csv', tname)
                self.full_paths.append(fpath)
                self.test_names.append(os.path.basename(fname))   
            
            # Initialize attributes
            self.data = pd.DataFrame()
            for path, name in zip(self.full_paths, self.test_names):
                df = pd.read_csv(path)
                # Force timestamp to be datetime
                df['name'] = name
                self.data = self.data.append(df)
            nrow, _ = self.data.shape
            self.data.index = np.arange(nrow)
        else:
            self.data, self.test_names, self.full_paths = evaluate.load_json_data(json_data)
          
        self.csv_dir = csv_dir
        # Read in a diagnostics csv path.  
        # Read csv, convert to dataframe
        self.data = pd.read_csv(self.csv_dir)
        rx_name = self.data.RX_Name
        self.rx_name = rx_name.to_numpy()
        a_weight = self.data.A_Weight
        self.a_weight = a_weight.to_numpy()
        fsf_all = self.data.FSF_Scores
        self.fsf_all = fsf_all.to_numpy()
        peak_dbfs = self.data.Peak_Amplitude
        self.peak_dbfs = peak_dbfs.to_numpy()
        self.trials = len(self.data) 
        aw_flag = self.data.AW_flag
        self.aw_flag = aw_flag.to_numpy()
        clip_flag = self.data.Clip_flag
        self.clip_flag = clip_flag.to_numpy()
        fsf_flag =self.data.FSF_flag
        self.fsf_flag = fsf_flag.to_numpy() 
        
    def to_json(self, filename=None):
        """
        Create json representation of diagnostics data

        Parameters
        ----------
        filename : str, optional
            If given save to json file. Otherwise returns json string. The default is None.

        Returns
        -------
        final_json: json
            json version of diagnotsic data and flag conditions

        """
        test_info = {}
            
        out_json = {
            'measurement': self.data.to_json(),
            'test_info': test_info
            }
            
        # Final json representation of all data
        final_json = json.dumps(out_json)
        if filename is not None:
            with open(filename, 'w') as f:
                json.dump(out_json, f)
                
        return final_json  
    
    @staticmethod
    def load_json_data(json_data):
        """
        Do all data loading from input json_data

        Parameters
        ----------
        json_data : TYPE
            DESCRIPTION.

        Returns
        -------
        test_names : list
            DESCRIPTION.
        test_paths : dict
            DESCRIPTION.
        data : pd.DataFrame
            DESCRIPTION.

        """  
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        # Extract data, cps, and test_info from json_data
        data = pd.read_json(json_data['measurement'])
        test_info = json_data['test_info']
        
        test_names = []
        test_paths = []
        for tname, tpath in test_info.items():
            test_names.append(tname)
            test_paths.append(tpath)
        
        
        # Return normal Access data attributes from these
        return data, test_names, test_paths 
 

            
    def fsf_plot(self):
        """
        Plot the FSF of every trial.  
        
        Returns
        -------
        None.
    
        """
        # Plot FSF values   
        x_axis = list(range(1,self.trials+1))
        dfFSF = pd.DataFrame({"FSF Score": self.fsf_all,
                              "Trial": x_axis,
                              "fsf_flag": self.fsf_flag})
        # Separate out flagged trials for easy plotting
        df_flag = dfFSF[dfFSF['fsf_flag'] == 1]
        fig = go.Figure()
        # plot all trials
        fig.add_trace(
            go.Scatter(
            x = dfFSF['Trial'],    
            y = dfFSF['FSF Score'],
            mode = 'markers',
            showlegend = True,
            name = 'FSF score',
            marker = dict(
                size = 10,
                color = '#0000FF',
            symbol = dfFSF.fsf_flag
                )
            )
        ), 
        # Plot all flagged trials
        fig.add_trace(
            go.Scatter(
            x = df_flag['Trial'],    
            y = df_flag['FSF Score'],
            mode = 'markers',
            showlegend = True,
            name = 'FSF score - flagged',
            marker = dict(
                size = 10,
                color ='red',
                symbol = 'square'
                )
            )
        ),        
        fig.update_layout(title_text='FSF Score of Received Audio')
        fig.update_xaxes(title_text='Trial Number')
        fig.update_yaxes(title_text='FSF Score')
        fig.show()      
        
    def aw_plot(self):
        """
        Plot the a-weight of every trial.    
        
        Returns
        -------
        None.
    
        """
        # Plot a-weighted power for all trials    
        x_axis = list(range(1,self.trials+1))
        dfAW = pd.DataFrame({"A-Weight": self.a_weight,
                             "Trial": x_axis,
                             "aw_flag":self.aw_flag})
        # Separate out flagged trials for easy plotting
        df_flag = dfAW[dfAW['aw_flag'] == 1]
        fig = go.Figure()
        # plot all trials
        fig.add_trace(
            go.Scatter(
                x = dfAW['Trial'], 
                y = dfAW['A-Weight'],
                mode = 'markers',
            showlegend = True,
            name = 'A-weight',
            marker = dict(
                size = 10,
                color = '#0000FF',
            symbol = dfAW.aw_flag
                )
            )
        ), 
        # Plot all flagged trials
        fig.add_trace(
            go.Scatter(
            x = df_flag['Trial'],    
            y = df_flag['A-Weight'],
            mode = 'markers',
            showlegend = True,
            name = 'A-weight - flagged',
            marker = dict(
                size = 10,
                color ='red',
                symbol = 'square'
                )
            )
        ),
        fig.update_layout(title_text='A-Weighted Power of Received Audio')
        fig.update_xaxes(title_text='Trial Number')
        fig.update_yaxes(title_text='A-Weight (dBA)')
        fig.show()
    
    def peak_dbfs_plot(self):
        """
        Plot the peak dbfs of every trial.    
        
        Returns
        -------
        None.
    
        """
        # Plot the peak amplitude (dbfs) for all trials   
        x_axis = list(range(1,self.trials+1))
        df_peak = pd.DataFrame({"Peak_dbfs": self.peak_dbfs,
                             "Trial": x_axis,
                             "clip_flag": self.clip_flag})  
        # Separate out flagged trials for easy plotting
        df_flag = df_peak[df_peak['clip_flag'] == 1]        
        fig = go.Figure()
        # Plot all trials
        fig.add_trace(
            go.Scatter(
                x = df_peak['Trial'], 
                y = df_peak['Peak_dbfs'],
                mode = 'markers',
                showlegend = True,
                name = 'Peak amplitude',
                marker = dict(
                size = 10,
                color = '#0000FF',
                symbol = df_peak.clip_flag
                )
            )
        ), 
        # Plot all flagged trials
        fig.add_trace(
            go.Scatter(
            x = df_flag['Trial'],    
            y = df_flag['Peak_dbfs'],
            mode = 'markers',
            showlegend = True,
            name = 'Peak amplitude - flagged',
            marker = dict(
                size = 10,
                color ='red',
                symbol = 'square'
                )
            )
        ),       
        fig.update_layout(title_text='Peak Amplitude of Received Audio')
        fig.update_xaxes(title_text='Trial Number')
        fig.update_yaxes(title_text='Peak Amplitude (dBfs)')
        fig.show()      


# -*- coding: utf-8 -*-
"""
Created on Tue Jan 12 08:43:22 2021

@author: jkp4
"""
import numpy as np


def bootstrap_ci(x, p=0.95, R=1e4, stat=np.mean, method="percentile"):
    """
    Bootstrap confidence interval for array of data.

    Parameters
    ----------
    x : np.ndarray
        Data for which to calculate a confidence interval for stat.
    p : float, optional
        Confidence level of output interval. The default is 0.95.
    R : int, optional
        Number of resamples to calculate. The default is 1e4.
    stat : function, optional
        Statistic for confidence interval. The default is np.mean.
    method : str, optional
        Defines the method used for bootstrapping. May be either
        "percentile" or "t". "percentile" performs a "quick-and-dirty"
        confidence interval, where "t" performs a bootstrap-t confidence
        interval.

    Returns
    -------
    ci : np.ndarray
         Confidence interval of data for stat

    resamples : list
         Values of stat for each resample

    Notes
    -----
    .. [1] Tim C. Hesterberg (2015) What Teachers Should Know About the
    Bootstrap: Resampling in the Undergraduate Statistics Curriculum, The
    American Statistician, 69:4, 371-386, DOI: 10.1080/00031305.2015.1089789

    """
    # Force R to be an int
    if type(R) is not int:
        R = int(R)
    # Length of data
    N = len(x)
    # Note observed statistic
    obs = stat(x)
    # Array to store resample stat values in
    resamples = []
    # Random generator
    gen = np.random.default_rng()

    if method == "percentile" or method == "p":
        # Generate resampled data, RxN matrix
        resamp = gen.choice(x, size=(R, N))
        # Calculate stat for each resample
        resamples = stat(resamp, axis=1)

        # Lower bound of confidence interval
        q_l = (1 - p) / 2
        # Upper bound of confidence interval
        q_u = 1 - q_l
        # Confidence interval
        ci = np.quantile(resamples, [q_l, q_u])

    # Standard error known only for sample mean, at the moment
    elif (method == "t") and (stat == np.mean):
        # Generate resampled data, RxN matrix
        resamp = gen.choice(x, size=(R, N))
        # Calculate stat for each resample
        rs_stat = stat(resamp, axis=1)
        # Calculate standard error for each resample
        rs_se = np.std(resamp, axis=1)/np.sqrt(N)
        # Estimate t-statistic for each resample
        rs_ts = (rs_stat - obs)/rs_se

        # Get t-score quantiles for esimating confidence interval
        q_l = (1 - p) / 2
        # Upper bound of confidence interval
        q_u = 1 - q_l
        bounds = np.quantile(rs_ts, [q_l, q_u])

        # Calculate CI from observed mean, observed standard error, and
        # t-scores from estimated t-distribution
        se = np.std(x)/np.sqrt(N)
        # Note that the bounds flip, see Ref [1] equation (4)
        ci = obs - bounds[::-1]*se

    elif (method == "t") and (stat != np.mean):
        raise ValueError("Standard error for this statistic is not implemented.")

    return (ci, resamples)


def approx_permutation_test(
    x, y, accept_threshold=0.05, R=1e4, stat=np.mean, tail="two"
):
    """
    Perform an approximate permutation test to test if two sets of data are
    from equivalent distributions.

    Null hypothesis in this test says that x is from the same distribution as
    y

    Parameters
    ----------
    x : NumPy Array
        Array of data from first condition

    y : NumPy Array
        Array of data from second condition

    accept_threshold : float
        Value that p-value is compared to either reject or accept the Null
        hypothesis. Default is 0.05.

    R : int
        Number of repetitions of resamples to perform. Default is 1e4

    stat : function
        Statistic to perform on data. Defaults to np.mean

    tail : str
        Determine if you do a one-sided or two-sided approximate permutation
        test. Acceptable values are "left", "right", or "two". If tail is
        "left", the test considers what fraction of resampled differences are
        less than the observed statistic. If tail is "right", the test
        considers what fraction of resampled differences are greater than the
        observed statistic. If tail is "two", the test considers what fraction
        of the absolute value of resampled differences are greater than the
        absolute value of the observed statistic.



    Return
    ------
    Boolean
        Rejection or not of the null hypothesis, in this case that data comes
        from same distribution. Explicitly if x and y are from distinct
        distributions this function will return True as the Null hypothesis
        assumes that they are from equivalent distributions. Alternatively if
        x and y are from the same distribution this function will return False.

    Examples
    --------
    Compare two extremely distinct data sets, returns True as the null
    hypothesis is rejected.

    >>> x = np.ones(120)
    >>> y = np.zeros(30)
    >>> mcvqoe.math.approx_permutation_test(x,y)
    True

    Compare two similar data sets, returns False as the null hypothesis is not
    rejected.

    >>> rng = np.random.default_rng()
    >>> x = rng.normal(0,1,100)
    >>> y = rng.normal(0.1,1,100)
    >>> mcvqoe.math.approx_permutation_test(x,y)
    False
    """

    # Force R to be an int
    if type(R) is not int:
        R = int(R)

    # Observed statistic
    observed = stat(x) - stat(y)

    # Number of trials in population 1
    m1 = len(x)
    # Number of trials in population 2
    m2 = len(y)

    # Total number of trials between populations
    n = m1 + m2
    # Combine both populations
    combo = np.concatenate((x, y))

    # Initialize difference array
    diffs = np.zeros((R, 1))

    # Random generator
    gen = np.random.default_rng()

    for k in range(R):
        # Create random permutation for resampling
        permIx = gen.permutation(n)

        # Grab first m elements of random order for population 1
        x_resamp = combo[permIx[0:m1]]
        # Grab remaining K elements of random order for population 2
        y_resamp = combo[permIx[m1:]]

        # Compute resample statistic (difference of means)
        diffs[k] = stat(x_resamp) - stat(y_resamp)

    # Calculate p-value for likelihood that observed came from mixed
    # distribution
    if tail == "left":
        # Consider how many resamples were lower than observed value (how far left is our observation?)
        pval_sum = diffs <= observed

    elif tail == "right":
        # Consider how many resamples were greater than observed value (how far right is our observation)
        pval_sum = diffs >= observed
    elif tail == "two":
        # Consider how many resamples were larger in magnitude than observed value (how far from either tale is our observation)
        pval_sum = abs(diffs) >= abs(observed)
    else:
        raise ValueError(
            "Unrecognized tail argument: {}. Must be either left, right, or two.".format(
                tail
            )
        )

    # Get fraction of resamples that achieved tail condtions
    pval = (np.count_nonzero(pval_sum)) / R

    if pval <= accept_threshold:
        # If pval is less than accept_threshold, reject Null hypothesis
        reject = True
    else:
        # If pval is greater than accept_threshold, do not reject Null hypothesis
        reject = False
    return reject


def standard_error(x):
    """
    Calculate standard error for data in x.

    This method assumes normality conditions are reasonable.

    Parameters
    ----------
    x : np.ndarray
        Data.

    Returns
    -------
    se : float
        Standard error of x

    """
    se = np.std(x) / np.sqrt(len(x))
    return se


def compare_uncs(x):
    """
    Calculate standard error and bootstrap based confidence intervals

    A nice sanity check that in well behaved circumstances, standard error is equivalent to bootstrap confidence interval results. Computes 95% confidence interval.

    Parameters
    ----------
    x : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    x_mean = np.mean(x)

    gum_u = standard_error(x)

    x_unc = 1.96 * gum_u

    print(
        "GUM Uncertainty (95% C.I.): x = {:.4f}, ({:.4f},{:.4f})".format(
            x_mean, x_mean - x_unc, x_mean + x_unc
        )
    )

    b_u, _ = bootstrap_ci(x)
    print(
        "Bootstrap Uncertainty (95% C.I.): x = {:.4f}, ({:.4f},{:.4f})".format(
            x_mean, b_u[0], b_u[1]
        )
    )


def improved_autocorrelation(x, lag_max=None):
    """
    Detect lags at which there is likely autocorrelation.

    Determined according to the improved bounds given in 'Zhang NF (2006)
    Calculation of the uncertainty of the mean of autocorrelated measurements'.

    Parameters
    ----------
    x : numpy array
        Numerical data on which to detect autocorrelation.

    Returns
    -------
    numpy array
        Array of indices for lags where this is likely autocorrelation.

    """
    # Force x to behave be a numpy array
    x = np.array(x)
    # Calculate sample autocorrelation estimate
    N = len(x)
    if lag_max is None:
        # TODO: Experiment more with the best lag_max to use
        # # Zhang NF (2006) recommends using a maximum lag of N/4
        lag_max = int(np.floor(N/4))
        
        # However that feels it can give some fairly noisey results, we instead use
        # the maximum lag used by R acf method
        # lag_max = int(np.floor(10*np.log10(N)))
    
    # Initialize correlation estimates
    corrs = np.zeros(lag_max)
    # Get mean of data
    m = np.mean(x)
    denom = np.sum(np.power(x-m, 2))
    for k in range(lag_max):
        numerator = np.sum((x[:(N-k)] - m) * (x[k:] - m))
        corrs[k] = numerator/denom
    
    # Respective uncertainties
    sigmas = np.zeros(lag_max)
    sigmas[0] = 1/np.sqrt(N)
    for k in range(1, lag_max):
        summer = np.sum(np.power(corrs[1:k], 2))
        nummer = 1 + 2* summer
        sigmas[k] = np.sqrt(nummer/N)
    

    return np.argwhere(np.abs(corrs) > 1.96 * sigmas)


def bootstrap_datasets_ci(*datasets, R=int(1e4), alpha=0.5):
    """
    Bootstrap for averaging means from different datasets.

    Parameters
    ----------
    *datasets : numpy arrays
        Datasets from which to take sample means. In context, the datasets
        are the different M2E sessions within a test.
    R : int, optional
        Number of resamples. The default is int(1e4).
    alpha : float, optional
        Alpha level of the test. The default is 0.5.

    Returns
    -------
    ci : numpy array
        Two element array containing the upper and lower confidence bound on
        the mean.

    """
    ds = datasets
    # TODO: No need to limit this to first dataset
    N = len(ds[0])
    x_bars = np.zeros((len(ds), R))
    for ii, dataset in enumerate(ds):
        rs = np.random.choice(dataset, size=(N, R))
        x_bar = np.mean(rs, axis=0)
        x_bars[ii, :] = x_bar
    
    # Means across sessions
    x_bar_dist = np.mean(x_bars, axis=0)
    # percentiles
    ql = alpha/2
    qu = 1 - ql
    ci = np.quantile(x_bar_dist, [ql, qu])
    return ci



import datetime
import importlib
import os
import platform
import shutil
import traceback
import warnings

import mcvqoe.base.version
import numpy as np


def fill_log(test_obj):
    """
    Take in QoE measurement class and fill in standard log entries

    ...

    Parameters
    ----------
    test_obj : QoE measurement class
        Class to generate test info for
    git_path : string, default=None
        path to git executable. Will look in the path if None
    """

    # initialize info
    info = {}

    # ---------------------------[RadioInterface info]---------------------------

    try:
        # Get ID and Version number from RadioInterface
        info["RI version"] = test_obj.ri.get_version()
        info["RI id"] = test_obj.ri.get_id()
    except AttributeError:
        # no RI for this object
        pass

    # ---------------------[Get traceback for calling info]---------------------

    # get a stack trace
    tb = traceback.extract_stack()

    # remove the last one cause that's this function
    tb = tb[:-1]

    # extract important info from traceback
    tb_info = [(os.path.basename(fs.filename), fs.name if fs.name != "<module>" else None) for fs in tb]

    # add entry for calling file
    info["filename"] = tb_info[-1][0]

    # format string with '->' between files
    info["traceback"] = "->".join([f"{f}({n})" if n is not None else f for f, n in tb_info])

    # ---------------------------[Add MCV QoE version]---------------------------

    info["mcvqoe version"] = mcvqoe.base.version

    # ----------------------[Add Measurement class version]----------------------

    if test_obj.__class__.__name__ == "measure":
        # get module for test_obj
        module = test_obj.__class__.__module__
    else:
        # TESTING : print base classes
        for base in test_obj.__class__.__bases__:
            # see if we have subclassed a measure class
            if base.__name__ == "measure":
                # get module from this class
                module = base.__module__
                # we are done
                break
        else:
            # could not find module
            module = None
            warnings.warn(f"Unable to find measure class for {test_obj.__class__.__name__}", category=RuntimeWarning)

    # set default version
    info["version"] = "Unknown"

    if module:
        # import base level module
        mod = importlib.import_module(module)
        try:
            info["version"] = mod.version
        except AttributeError as e:
            warnings.warn(f"Unable to get version {e}", category=RuntimeWarning)
            pass
    # ------------------------------[Add OS info]------------------------------

    info["os name"] = platform.system()
    info["os release"] = platform.release()
    info["os version"] = platform.version()

    # ---------------------------[Fill Arguments list]---------------------------

    # class properties to skip in all cases
    standard_skip = ["no_log", "info", "progress_update", "rng", "user_check"]
    arg_list = []

    np.set_string_function(lambda x: f"np.ndarray(dtype={x.dtype}, shape={x.shape})")
    for k, v in vars(test_obj).items():
        if not k.startswith("_") and k not in test_obj.no_log and k not in standard_skip:
            arg_list.append(k + " = " + repr(v))
    np.set_string_function(None)

    info["Arguments"] = ",".join(arg_list)

    # -------------------------------[Return info]-------------------------------

    return info


def format_text_block(text):
    """
    format text block for log.

    This writes out a, possibly, multi line text block to the log. It is used to
    write out both pre and post test notes.

    Parameters
    ----------
    text : str
        String containing the block of text to format.
    """

    if text is None:
        return ""

    return "".join(["\t" + line + "\n" for line in text.splitlines(keepends=False)])


def pre(info={}, outdir=""):
    """
    Take in M2E class info dictionary and write pre-test to tests.log.

    ...

    Parameters
    ----------
    info : dict
        The M2E.info dictionary used to write to tests.log.
    outdir : str
        The directory to write to.
    """

    # length to pad test params to
    pad_len = 10

    # Add 'outdir' to tests.log path
    log_datadir = os.path.join(outdir, "tests.log")

    skip_keys = ["test", "Tstart", "Pre Test Notes"]

    # Write all necessary arguments/test params into tests.log
    with open(log_datadir, "a") as file:

        file.write(f"\n>>{info['test']} started at {info['Tstart'].strftime('%d-%b-%Y %H:%M:%S')}\n")
        for key in info:
            if key not in skip_keys:
                file.write(f"\t{key:<{pad_len}} : {info[key]}\n")

        file.write("===Pre-Test Notes===\n")
        file.write(format_text_block(info["Pre Test Notes"]))


def post(info={}, outdir=""):
    """
    Take in M2E class info dictionary to write post-test to tests.log.

    ...

    Parameters
    ----------
    info : dict
        The M2E.info dictionary.
    outdir : str
        The directory to write to.

    """

    # Add 'outdir' to tests.log path
    log_datadir = os.path.join(outdir, "tests.log")

    with open(log_datadir, "a") as file:
        if "Error Notes" in info:
            notes = info["Error Notes"]
            header = "===Test-Error Notes==="
        else:
            header = "===Post-Test Notes==="
            notes = info.get("Post Test Notes", "")

        # write header
        file.write(header + "\n")
        # write notes
        file.write(format_text_block(notes))
        # write end
        file.write("===End Test===\n\n")



import os.path
import shutil
import subprocess
import sys
import tempfile
import warnings

import mcvqoe.base
import numpy as np
import scipy.signal
from mcvqoe.base import audio_float
from mcvqoe.delay.ITS_delay import active_speech_level

try:
    # try to import importlib.metadata
    from importlib.metadata import entry_points
except ModuleNotFoundError:
    # fall back to importlib_metadata
    from importlib_metadata import entry_points


def _get_entry_points(group):
    if sys.version_info >= (3, 10):
        #use new selection mechanism
        return entry_points(group=group)
    else:
        #use the old way
        return entry_points()[group]


class QoEsim:
    """
    Class to use for simulations.

    This class simulates the functionalities of mcvqoe.hardware.AudioPlayer and
    mcvqoe.hardware.RadioInterface.

    Parameters
    ----------
    port : str, default=None
        For compatibility with 'RadioInterface'. Value has no effect on simulation.

    Attributes
    ----------
    debug : bool, default=False
        If true, print some extra 'RadioInterface' things.
    sample_rate : int, default=48000
        Sample rate of audio in/out in samples per second.
    blocksize : int, default=512
        For compatibility with 'AudioPlayer'. Value has no effect on simulation.
    buffersize : int, default=20
        For compatibility with 'AudioPlayer'. Value has no effect on simulation.
    overplay : float, default =1.0
        The number of seconds of extra audio to play/record at the end of a clip.
    device : str, default=__class__
        Device name for compatibility with 'AudioPlayer'. This is set to the
        string representation of the class instance by default. Changing has no
        effect on simulations.
    port_name : str, default='SIM'
        Serial port name. For compatibility with RadioInterface, changing has no
        effect on simulations.
    rec_chans : dict
        Dictionary describing the recording. Dictionary keys must be one of
        {'rx_voice','PTT_signal'. For simulation, the value is ignored, as this
        normally represents the physical channel number. At this time QoEsim can not
        simulate 'timecode' or 'tx_beep' audio.
    playback_chans : dict
        Dictionary describing the playback channels. Dictionary keys should be
        one of {'tx_voice','start_signal'}. For simulation, the value is ignored,
        as this normally represents the physical channel number.
    channel_tech : string, default='clean'
        Technology to use for the simulated channel. Must match the name of an
        installed channel plugin.
    channel_rate : int or str, default=None
        Rate to simulate channel at. Each channel tech handles this differently.
        When set to None the default rate is used.
    pre_impairment : function, default=None
        Impairment function to apply before the audio goes through the channel.
        This function is passed an audio vector and should return a audio vector,
        of the same length, with the impairments applied. A value of None skips
        applying an impairment.
    post_impairment : function, default=None
        Impairment function to apply after the audio goes through the channel.
        This function is passed an audio vector and should return a audio vector,
        of the same length, with the impairments applied. A value of None skips
        applying an impairment.
    channel_impairment : function, default=None
        Imparment to apply to the channel data. Callback function that is called
        on the channel data to simulate a less than perfect communication channel.
        A value of none skips applying an impairment.
    m2e_latency : float, default=None
        Simulated mouth to ear latency for the channel in seconds. If none than
        the minimum latency for the technology is used. Can be set to a callable
        object to return a different value each time.
    access_delay : float, default=0
        Delay between the time that the simulated push to talk button is pushed
        and when audio starts coming through the channel. If the 'ptt_delay'
        method is called before play_record is called, then the time given to
        'ptt_delay' is added to access_delay to get the time when access is
        granted. Otherwise access is granted 'access_delay' seconds after the
        clip starts.Can be set to a callable object to return a diffrent value
        each time. Can be set to a callable object to return a different value
        each time.
    device_delay : float, default=0
        Delay of simulated audio interface. This is added to m2e_latency and is
        the delay for the PTT signal. Can be set to a callable object to return
        a different value each time.
    rec_snr : float, default=60
        Signal to noise ratio, in dB, for audio channel.
    print_args : bool, default=False
        Print arguments to external programs.
    PTT_sig_freq : float, default=409.6
        Frequency of the PTT signal from the play_record method.
    PTT_sig_amplitude : float, default=0.7
        Amplitude of the PTT signal from the play_record method.
    default_radio : int, default=1
        For compatibility with RadioInterface, the number of the radio to use
        when none is given. Doesn't really have an effect on simulation, but the
        ptt_wait_delay is tracked for each radio.
    allow_all_rec : bool, default=False
        Accept all recording channels, even if we don't know how to generate
        data for them. If an unexpected channel is found, it will be filled with
        random data. This is usefull for basic functionality tests for two
        location.

    See Also
    --------
    mcvqoe.hardware.AudioPlayer : Play audio on real hardware.
    mcvqoe.hardware.RadioInterface : Key real radios.

    Examples
    --------
    Play 48 kHz audio stored in tx_voice and record in a file named 'test.wav'.

    >>> import mcvqoe.simulation.QoEsim
    >>> sim_obj=mcvqoe.simulation.QoEsim(fs=int(48e3))
    >>> sim_obj.play_record(tx_voice,'test.wav')

    Now do the same but also output the start signal on channel 1 and record the
    PTT signal on channel 1.

    >>> sim_obj.playback_chans={'tx_voice':0,'start_signal':1}
    >>> sim_obj.rec_chans={'rx_voice':0,'PTT_signal':1}
    >>> sim_obj.play_record(tx_voice,'test.wav')
    """

    def __init__(
        self,
        port=None,
        **kwargs
    ):

        #values for RadioInterface
        #NOTE : port, is added as an argument for compatibility, but is not used.
        self.debug = False
        self.PTT_state = [
            False,
        ] * 2
        self.LED_state = [
            False,
        ] * 2
        self.ptt_wait_delay = [
            -1.0,
        ] * 2
        # values for AudioPlayer
        self.sample_rate = int(48e3)
        self.blocksize = 512
        self.buffersize = 20
        self.overplay = 1.0
        self.device = str(__class__)  # fake device name
        self.port_name = 'SIM'   # fake serial port
        self.rec_chans = {"rx_voice": 0}
        self.playback_chans = {"tx_voice": 0}
        # channel variables
        self.channel_tech = "clean"
        self.channel_rate = None
        self.pre_impairment = None
        self.post_impairment = None
        self.channel_impairment = None
        # set to none to get min delay
        self.m2e_latency = None
        self.access_delay = 0
        self.device_delay = 0
        # SNR for audio in dB
        self.rec_snr = 60
        # print arguments sent to external programs for debugging
        self.print_args = False
        # PTT signal parameters
        self.PTT_sig_freq = 409.6  # TODO : VERIFY!
        self.PTT_sig_amplitude = 0.7
        self.default_radio = 1

        #allow any channel type to be recorded, fill with random data, if needed
        self.allow_all_rec = False

        #get properties from kwargs
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            #check if the 'fs' argument was given
            elif k == 'fs':
                #fs gets translated to sample_rate for legacy purposes
                self.sample_rate = v
            else:
                raise TypeError(f"{k} is not a valid keyword argument")

    def __repr__(self):
        string_props=('sample_rate','overplay','rec_chans','playback_chans','channel_tech','channel_rate','m2e_latency','access_delay','device_delay','rec_snr')
        string_opt_props=('pre_impairment','post_impairment','channel_impairment')

        props=[]

        for prop in string_props:
            props.append(f'{prop} = {repr(getattr(self, prop))}')
        for prop in string_opt_props:
            if getattr(self, prop) is not None:
                props.append(f'{prop} = {repr(getattr(self, prop))}')

        return f'{type(self).__name__}({", ".join(props)})'

    def __enter__(self):

        return self

    def ptt(self, state, num=None):
        """
        Change the push-to-talk status of the radio interface.

        For 'RadioInterface' this would turn on or off the PTT outputs. For simulation,
        This function does not have much of an effect. It will clear any PTT time set by
        the 'ptt_delay' method.

        Parameters
        ----------
        state : bool
            State to set PTT output to.
        num : int, default=None
            PTT output number to use. If None, self.default_radio is used.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.ptt : Same function but, for hardware.

        Examples
        --------
        Key a fake radio.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> sim_obj.ptt(True)  #key radio
        >>> sim_obj.ptt(False) #de-key radio
        """

        if num is None:
            num = self.default_radio

        self.PTT_state[num] = bool(state)
        # clear wait delay
        self.ptt_wait_delay[num] = -1

    def led(self, num, state):
        """
        Turn on and off fake LEDs.

        For 'RadioInterface' this would turn on or off LEDs on the device. For simulation
        this function doesn't do much.

        Parameters
        ----------
        num : int
            LED number to set.
        state : bool
            The LED state to set. True is on, False is off.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.led : Same function but, for hardware.

        Examples
        --------
        Turn on some fake LEDs.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> sim_obj.led(1,True)
        >>> sim_obj.led(2,True)
        >>> sim_obj.led(1,False)
        """

        # check LED state
        if state:
            if self.debug:
                print("RadioInterface LED {%num} state is on")
        else:
            if self.debug:
                print("RadioInterface LED {%num} state is off")

        # set state in simulation object
        self.LED_state[num-1] = bool(state)

    def devtype(self):
        """
        Get the devicetype string from the fake radio interface.

        For 'RadioInterface' this would query the firmware version of the board.
        For simulation, this just returns a version string.

        Returns
        -------
        str
            A string pretending to come from a 'RadioInterface' board.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.devtype : Same function but, for hardware.

        Examples
        --------
        Query a fake 'RadioInterface'.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> print(sim_obj.devtype())
        """

        # TODO : simulate other versions??
        return "MCV radio interface : v1.2.1"

    def get_id(self):
        """
        Get the ID string from radio interface.

        For 'RadioInterface' this would query the firmware version of the board.
        For simulation, this just returns the filename.

        Returns
        -------
        str
            A string pretending to come from a 'RadioInterface' board.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.get_id : Same function but, for hardware.

        Examples
        --------
        Query a fake 'RadioInterface'.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> print(sim_obj.get_id())
        """

        return os.path.basename(__file__)

    def get_version(self):
        """
        Get the version string from fake 'RadioInterface'.

        For 'RadioInterface' this would query the firmware version of the board.
        For simulation, this just returns the mcvqoe library version.

        Returns
        -------
        str
            A string pretending to come from a 'RadioInterface' board.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.get_version : Same function but, for hardware.

        Examples
        --------
        Query a fake 'RadioInterface'.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> print(sim_obj.get_version())
        """

        return mcvqoe.base.version

    def pttState(self):
        """
        Get the PTT state from fake 'RadioInterface'.

        Return the state of the simulated PTT outputs.

        Returns
        -------
        bool array
            An array with the status of each simulated PTT output.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.pttState : Same function but, for hardware.

        Examples
        --------
        Query a fake 'RadioInterface'.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> print(sim_obj.pttState())
        >>> sim_obj.ptt(True)
        >>> print(sim_obj.pttState())
        """
        return self.PTT_state

    def waitState(self):
        """
        Get the PTT wait state from fake 'RadioInterface'.

        For 'RadioInterface' this would check if PTT outputs were waiting for a
        start signal. For simulation this always returns 'Idle'

        Returns
        -------
        str
            This function always returns 'Idle'.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.waitState : Same function but, for hardware.

        Examples
        --------
        Query a fake 'RadioInterface'.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> print(sim_obj.waitState())
        """

        # TODO: do we need to simulate this better?
        return "Idle"

    def ptt_delay(self, delay, num=None, use_signal=False):
        """
        Set how far into the clip to key the fake radio.

        For 'RadioInterface' this would setup the PTT output to key up with a
        specified delay. For simulation, this sets the time in the clip that
        'access_delay' is referenced to.

        Parameters
        ----------
        delay : float
            The number of seconds between the start of the clip and when the
            'access_delay' time starts counting.
        num : int, default=None
            The PTT output to use. If None, use self.default_radio.
        use_signal : bool, default=False
            For 'RadioInterface', this would control if the start signal is used
            to reference the 'delay' to the beginning of the clip. For simulation
            'delay' is always referenced to the beginning of the clip.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.ptt_delay : Same function but, for hardware.
        mcvqoe.simulation.QoEsim.play_record : Where 'delay' times are used.

        Examples
        --------
        Generate a simulated clip where the radio is keyed up 1 second into the
        clip. 'tx_voice' has the 48k Hz voice vector. The result is stored in
        'test.wav'

        >>> sim_obj=mcvqoe.simulation.QoEsim(fs=int(48e3))
        >>> sim_obj.playback_chans={'tx_voice':0,'start_signal':1}
        >>> sim_obj.rec_chans={'rx_voice':0,'PTT_signal':1}
        >>> sim_obj.ptt_delay(1)
        >>> sim_obj.play_record(tx_voice,'test.wav')
        """

        if num is None:
            num = self.default_radio

        self.ptt_wait_delay[num] = delay
        # set state to true, this isn't 100% correct but the delay will be used
        # for the sim so, it shouldn't matter
        self.PTT_state[num] = True

    def temp(self):
        """
        Read fake temperatures from fake hardware.

        For 'RadioInterface' this would read temperature from the microcontroller.
        For simulation, this returns some fake values.
        NOTE : the current hardware no longer supports temperature measurements,
        this may be removed in the future.

        Returns
        -------
        int array
            Two integers representing the internal temperature and external
            temperature.

        See Also
        --------
        mcvqoe.hardware.RadioInterface.temp : Same function but, for hardware.

        Examples
        --------
        Read some fake temperatures.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> print(sim_obj.temp())
        """
        # TODO : generate better fake values??
        return (38, 1500)

    # delete method
    def __del__(self):
        # nothing to do here
        pass

    def __exit__(self, exc_type, exc_value, exc_traceback):

        self.ptt(False)
        self.led(1, False)

    # =====================[Find device function]=====================
    @staticmethod
    def find_device(device_str="UMC"):
        """
        Return fake device name.

        For 'AudioPlayer' this would return the name of the audio device to use.
        For simulation this just returns the string representation of the class.

        Arguments
        ---------
        device_str : str, default="UMC"
            For compatibility with AudioPlayer. For hardware this would look for
            device matching this. For simulations, it does nothing.

        Returns
        -------
        str
            The string representation of the class instance.

        See Also
        --------
        mcvqoe.hardware.AudioPlayer.find_device : Same function but for hardware.

        Examples
        --------
        Query a fake 'AudioPlayer'.

        >>> sim_obj=mcvqoe.simulation.QoEsim()
        >>> print(sim_obj.find_device())
        """
        return str(__class__)

    # =====================[Get Channel Technologies]=====================
    @staticmethod
    def get_channel_techs():
        """
        Return a list of channel technologies.

        This returns a list of the channel technologies that have been found.
        These can be used as values for the `channel_tech` property.

        Returns
        -------
        list of strings
            A list of the names of channel technologies.

        See Also
        --------
        mcvqoe.simulation.QoEsim.channel_tech : Channel technology to use.
        mcvqoe.simulation.QoEsim.play_record : Function to simulate a channel.

        Examples
        --------

        List channel technologies.

        >>>QoEsim.get_channel_techs()
        ("clean")
        """
        chan_types = []
        for c in _get_entry_points("mcvqoe.channel"):
            chan_types.append(c.name)

        return tuple(chan_types)

    # =====================[Get Channel Rates]=====================
    @staticmethod
    def get_channel_rates(tech):
        """
        Return possible rates for a given channel tech.

        This queries a channel plugin for the default rate and a list of unique
        rates that can be used on the channel.

        Returns
        -------
        default_rate : string or number
            Default rate for channel, if no rate is used, None is returned.
        rates : list
            A list of the names of channel technologies. If no rate is used, an
            empty list is returned.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_channel_techs : List of channel technologies.
        mcvqoe.simulation.QoEsim.channel_tech : Channel technology to use.
        mcvqoe.simulation.QoEsim.play_record : Function to simulate a channel.

        Examples
        --------
        Get rates for a clean channel

        >>> QoEsim.get_channel_techs('clean')
        (None,[])
        """
        mod = QoEsim._get_chan_mod(tech)
        return (mod.default_rate, mod.rates)

    # =====================[Get Channel Version]=====================
    @staticmethod
    def get_channel_version(tech):
        """
        Return version string for channel tech.

        This queries a channel plugin for the installed version.

        Returns
        -------
        string
            Channel version string.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_channel_techs : List of channel technologies.
        mcvqoe.simulation.QoEsim.channel_tech : Channel technology to use.
        mcvqoe.simulation.QoEsim.play_record : Function to simulate a channel.

        Examples
        --------
        Get version for a clean channel

        >>> QoEsim.get_channel_version('clean')
        <mcvqoe.base.version>
        """
        mod = QoEsim._get_chan_mod(tech)

        try:
            ver = mod.version
        except AttributeError:
            ver = 'Unknown'

        return ver

    # =====================[Get Channel Version]=====================
    @staticmethod
    def get_channel_type(tech):
        """
        Return channel type for channel tech.

        This queries a channel plugin for the channel type. Right now there are
        three defined channel types 'digital-channel', 'analog-channel' and
        'audio'.

        Digital channels should pass a boolean array to the channel impairment
        function. Where each element represents one bit in the stream.

        Analog channels should pass a floating point array to the channel
        impairment function. Where each element represents the volage vs time in
        the channel.

        Audio channels are just audio channel data. Audio channels should pass a
        floating point array to the channel impairment function, where each
        sample represents voltage over time.

        Returns
        -------
        string
            Channel type string.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_channel_techs : List of channel technologies.
        mcvqoe.simulation.QoEsim.channel_tech : Channel technology to use.
        mcvqoe.simulation.QoEsim.play_record : Function to simulate a channel.

        Examples
        --------
        Get version for a clean channel

        >>> QoEsim.get_channel_type('clean')
        'audio'
        """
        mod = QoEsim._get_chan_mod(tech)

        try:
            chan_type = mod.channel_type
        except AttributeError:
            chan_type = 'Unknown'

        return chan_type
    # =====================[get channel module]=====================
    @staticmethod
    def _get_chan_mod(tech):
        '''
        Get module for channel plugin.
        '''

        chan_types = []
        # locate any channel plugins installed
        for c in _get_entry_points("mcvqoe.channel"):
            if c.name == tech:
                #load module and return
                return c.load()
            chan_types.append(c.name)
        #tech not found, raise error
        raise ValueError(f'Unknown channel tech "{tech}" valid channel technologies are {chan_types}')

    # =====================[get impairment plugins]=====================
    @staticmethod
    def _get_impairments():
        '''
        Get installed impairments.

        Return a tuple of EntryPoint objects for the installed impairments.

        Returns
        -------
        tuple
            Tuple with one EntryPoint object for each impairment plugin.
        '''
        # locate any impairment plugins installed
        return _get_entry_points("mcvqoe.impairment")

    # =========================[get impairment plugins]=========================
    def _get_impairment_module(name):
        """
        Get matching impairment module.
        """

        # locate any impairment plugins installed
        impairments = QoEsim._get_impairments()

        for i in impairments:
            if i.name == name:
                module = i.load()
                break
        else:
            raise ValueError(f'impairment name \'{name}\' is not installed')

        return module
    # =========================[get impairment plugins]=========================
    @staticmethod
    def get_impairment_names(plugin_type):
        """
        Return the names of all the impairments that match a given type.

        This imports all the impairment plugins, imports them and checks to see
        if their impairment type matches `plugin_type`.

        Parameters
        ----------
        plugin_type : str
            A string containing the plugin type to match. Should be one of :
            'Analog', 'Digital' or 'Audio'.

        Returns
        -------
        tuple
            A tuple containing the names of the plugins that match the given type.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_channel_type : Get the type of a channel.
        mcvqoe.simulation.QoEsim.get_channel_techs : List of channel technologies.
        mcvqoe.simulation.QoEsim.channel_tech : Channel technology to use.
        """

        # locate any impairment plugins installed
        impairments = QoEsim._get_impairments()

        names = []

        for i in impairments:
            #TODO : probably need a try in her somewhere
            #load module
            module = i.load()
            #check if type matches
            if plugin_type == module.impairment_type or 'generic' == module.impairment_type:
                names.append(i.name)

        return tuple(names)

    # =======================[get all impairment plugins]=======================
    @staticmethod
    def get_all_impairment_names():
        """
        Return the names of all the impairments that are installed.

        Returns
        -------
        tuple
            A tuple containing the names of the plugins that match the given type.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_impairment_names : Get matching impairments.
        """

        # locate any impairment plugins installed
        impairments = QoEsim._get_impairments()

        return tuple([i.name for i in impairments])

    # =======================[get impairment parameters]=======================
    @staticmethod
    def get_impairment_params(name):
        """
        Get the parameters for a given impairment.

        Returns a list containing an ImpairmentParam object for each impairment
        that the `create_impairment` function in the plugin takes.

        Returns
        -------
        list
            A list of ImpairmentParam objects describing the parameters.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_impairment_func : Create impairment function.
        """
        module = QoEsim._get_impairment_module(name)

        return module.parameters

    # =======================[get impairment Description]=======================
    @staticmethod
    def get_impairment_description(name):
        """
        Get the description from an impairment plugin.

        Parameters
        ----------
        name : str
            A string containing the impairment to get the description of.

        Returns
        -------
        str
            The description from the plugin.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_impairment_names : Get matching impairments.
        """
        module = QoEsim._get_impairment_module(name)

        return module.description

    # =======================[get impairment Version]=======================
    @staticmethod
    def get_impairment_version(name):
        """
        Get the version string of an impairment plugin

        Parameters
        ----------
        name : str
            A string containing the impairment to get the version of.

        Returns
        -------
        str
            The version string from the plugin.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_impairment_names : Get matching impairments.
        """


        module = QoEsim._get_impairment_module(name)

        try:
            ver = module.version
        except AttributeError:
            ver = 'Unknown'

        return ver

    # ========================[initialize an impairment]========================
    @staticmethod
    def get_impairment_func(name, **kwargs):
        """
        Return an impairment function.

        Return an impairment function for the named impairment with the given
        arguments. Possible keyword arguments are given by get_impairment_params.
        Arguments that are not given will have their default values used. Extra
        arguments that are given will be ignored.

        Parameters
        ----------
        name : str
            The plugin to create an impairment for.

        Returns
        -------
        object
            A callable object that can be used as an impairment for QoEsim.

        See Also
        --------
        mcvqoe.simulation.QoEsim.get_impairment_params : Get impairment parameters.
        """
        module = QoEsim._get_impairment_module(name)

        return module.create_impairment(**kwargs)

    # ========================[get delay value]========================
    def _get_delay_samples(self, value, name, offset=0):
        if value is None:
            delay_samples = 0
        elif callable(value):
            # correct for offset
            delay_samples = int((value() - offset) * self.sample_rate)
            #fix delay to zero
            if delay_samples<0:
                #give warning about change
                warnings.warn(f'{name} would be {delay_samples} samples, replacing with 0')
                #set to zero
                delay_samples = 0
        else:
            # correct for offset
            delay_samples = int((value - offset) * self.sample_rate)

        if delay_samples < 0:
            raise ValueError(
                f"Unable to simulate {name} of {value}. Minimum simulated {name} for technology '{self.channel_tech}' is {offset}"
            )

        return delay_samples
    # =========================[record audio function]=========================
    def play_record(self, audio, out_name):
        """
        Simulate playing and recording of audio through a communication channel.

        This function simulates a communication channel. The values in
        'self.playback_chans' are checked for valid values, but they don't really
        have any other effect. Only 'rech_chans' of type 'rx_voice' and
        'PTT_signal' can be used.

        Parameters
        ----------
        audio : array
            An array of audio data to pass through the communication channel.
        out_name : str
            The name of the output .wav file to write to.

        Returns
        -------
        list of strings
            A list of the recorded output channels in the order that they appear
            in the output file.

        See Also
        --------
        mcvqoe.hardware.AudioPlayer.play_record : Same function but for hardware.

        Examples
        --------
        Play 48 kHz audio stored in tx_voice and record in a file named
        'test.wav'.

        >>> import mcvqoe.simulation.QoEsim
        >>> sim_obj=mcvqoe.simulation.QoEsim(fs=int(48e3))
        >>> sim_obj.play_record(tx_voice,'test.wav')

        Now do the same but also output the start signal on channel 1 and record
        the PTT signal on channel 1.

        >>> sim_obj.playback_chans={'tx_voice':0,'start_signal':1}
        >>> sim_obj.rec_chans={'rx_voice':0,'PTT_signal':1}
        >>> sim_obj.play_record(tx_voice,'test.wav')
        """
        # loop through playback_chans this is mostly just a format check to make
        # sure that all keys used are valid
        for (k, v) in self.playback_chans.items():
            if k == "start_signal":
                pass
            elif k == "tx_voice":
                pass
            else:
                raise ValueError(f"Unknown output channel : {k}")

        # list of outputs that we need to produce
        outputs = []
        # loop through rec_chans and detect keys that we can't produce
        for (k, v) in self.rec_chans.items():
            if k not in ("rx_voice", "PTT_signal"):
                if self.allow_all_rec:
                    warnings.warn(f"Got unexpected recording type : '{k}'")
                else:
                    raise RuntimeError(f"{__class__} can not generate recordings of type '{k}'")
            outputs.append(k)

        # get the module for the chosen channel tech
        chan_mod = self._get_chan_mod(self.channel_tech)

        # get delay offset for channel technology
        m2e_offset = chan_mod.standard_delay

        # calculate values in samples
        overplay_samples = int(self.overplay * self.sample_rate)

        # get value for device delay
        device_delay_samples = self._get_delay_samples(self.device_delay,'device delay')

        # get value for M2E
        m2e_latency_samples = self._get_delay_samples(self.m2e_latency,'M2E',offset = m2e_offset)

        # convert audio values to floats to work on them
        float_audio = mcvqoe.base.audio_float(audio)

        # append overplay to audio
        overplay_audio = np.zeros(int(overplay_samples), dtype=np.float32)
        tx_data_with_overplay = np.concatenate((float_audio, overplay_audio))


        # check if PTT was keyed during audio
        if self.ptt_wait_delay[self.default_radio] == -1:
            # PTT wait not set, don't simulate access time
            ptt_st_dly_samples = 0
            access_delay_samples = 0
        else:
            access_delay_samples = self._get_delay_samples(self.access_delay,'Access Delay')
            ptt_st_dly_samples = int(self.ptt_wait_delay[self.default_radio] * self.sample_rate)

        # mute portion of tx_data that occurs prior to triggering of PTT
        muted_samples = int(access_delay_samples + ptt_st_dly_samples)
        muted_tx_data_with_overplay = tx_data_with_overplay[muted_samples:]

        # add pre channel impairments
        if self.pre_impairment:
            muted_tx_data_with_overplay = self.pre_impairment(
                muted_tx_data_with_overplay, self.sample_rate
            )

        if self.rec_snr is None:
            # don't add any noise
            muted_tx_data_with_overplay_and_noise = muted_tx_data_with_overplay
        else:
            # generate gaussian noise, of unit standard deveation
            noise = np.random.normal(0, 1, len(muted_tx_data_with_overplay)).astype(np.float32)

            # measure amplitude of signal and noise
            sig_level = active_speech_level(tx_data_with_overplay, self.sample_rate)
            noise_level = active_speech_level(noise, self.sample_rate)

            # calculate noise gain required to get desired SNR
            noise_gain = sig_level - (self.rec_snr + noise_level)

            # set noise to the correct level
            noise_scaled = noise * (10 ** (noise_gain / 20))

            # add noise to audio
            muted_tx_data_with_overplay_and_noise = muted_tx_data_with_overplay + noise_scaled

        # call audio channel function from module
        channel_voice = chan_mod.simulate_audio_channel(
            muted_tx_data_with_overplay_and_noise,
            self.sample_rate,
            self.channel_rate,
            self.print_args,
            self.channel_impairment,
        )

        # add post channel impairments
        if self.post_impairment:
            channel_voice = self.post_impairment(channel_voice, self.sample_rate)

        # generate silent noise section comprised of ptt_st_dly, access delay
        # and m2e latency audio snippets
        silence_length = int(ptt_st_dly_samples
                             + access_delay_samples
                             + m2e_latency_samples
                             + device_delay_samples)
        silent_section = np.zeros(silence_length)
        # prepend silent section to rx_data
        rx_voice = np.concatenate((silent_section, channel_voice))
        # Derive mean and standard deviation from real-world noise observed in
        # the audio recordings. This basically simulates the noise floor from
        # audio cables/audio interface that we always have when we record.
        mean = 0
        std = 1.81e-5
        rx_noise = np.random.normal(mean, std, len(rx_voice))
        # Add noise to rx voice
        rx_voice = rx_voice + rx_noise

        # force rx_data to be the same length as tx_data_with_overplay
        rx_voice = rx_voice[: tx_data_with_overplay.shape[0]]

        # output array is the length of rx_voice x number of outputs
        rx_data = np.empty((rx_voice.shape[0], len(outputs)))

        for n, o_type in enumerate(outputs):
            if o_type == "PTT_signal":
                # calculate PTT tone start time
                ptt_tone_st_samples = ptt_st_dly_samples + device_delay_samples
                # calculate length of sine signal in samples
                sin_len_samples = rx_data.shape[0] - ptt_tone_st_samples
                # construct sine signal
                rx_data[ptt_tone_st_samples:, n] = self.PTT_sig_amplitude * np.sin(
                    2
                    * np.pi
                    * self.PTT_sig_freq
                    * np.arange(sin_len_samples)
                    / float(self.sample_rate)
                )
                # zero out before PTT
                rx_data[:ptt_tone_st_samples, n] = 0
            elif o_type == "rx_voice":
                # add data to the array
                rx_data[:, n] = rx_voice
            #fallback if allow_all_rec is true
            elif self.allow_all_rec:
                #don't know how to generate this signal, fill with noise
                rx_data[:, n] = np.random.normal((rx_voice.shape[0],))
            else:
                raise RuntimeError("Internal error")

        # write out audio file
        mcvqoe.base.audio_write(out_name, int(self.sample_rate), rx_data)

        return outputs

    def record(self, filename):
        """
        Record audio based on rec_chans.

        There is no connection to the Tx side, so fixed length noise will be
        recorded.

        Parameters
        ----------
        filename : str
            The file name to write audio to.

        """

        if not self.allow_all_rec:
            raise RuntimeError("Data can not be generated for record.")

        samples = 10 * int(self.sample_rate)

        #generate data
        rx_data = np.random.normal((samples,len(self.rec_chans)))

        # write out audio file
        mcvqoe.base.audio_write(filename, int(self.sample_rate), rx_data)

        #return a tuple of channels
        return tuple(self.rec_chans.keys())

class ImpairmentParam:
    '''
    Class for defining parameters to impairments.
    '''
    def __init__(self, default, value_type, choice_type, **kwargs):
        self.default = default
        self.value_type = value_type
        self.choice_type = choice_type

        if self.choice_type == 'positive':
            self.max_val = np.inf
            self.min_val = 0.0 + kwargs['interval']

        #add kwargs
        for k, v in kwargs.items():
            #set attribute, don't care if it exists
            setattr(self, k, v)

class Impairment:
    '''
    Class used to define impairments.

    This is a good base class to make impairments that will be logged correctly.
    '''
    def __init__(self, **kwargs):
        self.arguments =  kwargs

    def __repr__(self):
        return f'{type(self).__name__}('+', '.join([f'{k}={repr(v)}' for k, v in self.arguments.items()])+')'

    def __call__(self, data, fs):
        return self.impairment(data, fs, **self.arguments)


#!/usr/bin/env python

import argparse
import mcvqoe.simulation

def main():
    parser = argparse.ArgumentParser(
            description='Get info on mcvqoe channel plugins. If a plugin name' +
                        ' is given, print info on that plugin. Otherwise list' +
                        ' available plugins.')
    parser.add_argument(
    'chan', default=None, nargs="?", metavar='CHAN',
    help='audio channel to print info for.')
    
    args = parser.parse_args()
    
    #create qoesim object
    sim_obj=mcvqoe.simulation.QoEsim()
    #get channel technologies
    techs=sim_obj.get_channel_techs()
    #check what info we are going to print
    if(args.chan):
        #check that this tech is valid
        if(args.chan not in techs):
            print(f'Error : {args.chan} is not a valid channel technology')
            exit(1)

        mod = sim_obj._get_chan_mod(args.chan)
        print(f'module name : {mod.__name__}')

        ver = sim_obj.get_channel_version(args.chan)
        print(f'{args.chan} channel version : {ver}')

        (def_rate,rates)=sim_obj.get_channel_rates(args.chan)
        print(f'Default rate : {def_rate}')
        print(f'Possible rates : {", ".join(rates)}')

        #everything is fine
        exit(0)
    else:
        print(f'Channel technologies : {", ".join(techs)}')

        #everything is fine
        exit(0)


if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 26 11:55:08 2021

@author: jkp4
"""
import numpy as np
import resample_techniques as rt
import mcvqoe.math
mu = 0
sig = 1
rng = np.random.default_rng()
N_tests = 1000
N = 50

results = []
for k in range(N_tests):
    if k % 50 == 0:
        print(f'{100*k/N_tests}% done')
    sample = rng.normal(mu, sig, N)
    obs = np.mean(sample)
    # ci, _ = rt.bootstrap_ci(sample)
    ci, _ = mcvqoe.math.bootstrap_ci(sample, method="p", R=1e4)
    if ci[0] <= mu and mu <= ci[1]:
        results.append(True)
    else:
        results.append(False)

print(f'Percent success: {np.mean(results)}')



from mcvqoe.base import audio_float

import numpy as np
import datetime

soft_time_fmt = "TM%j-%Y_%H-%M-%S.%f"


def soft_time_decode(audio,fs):
    if len(audio.shape) != 1:
        raise ValueError("Input must be a numpy vector")
    
    # Get prefix from time format
    prefix = soft_time_fmt[: soft_time_fmt.index("%")]

    # TODO : raise error if prefix is zero length???

    # convert to integer values
    audio = np.round(audio_float(audio) * 255.0).astype(int)

    # convert sequence to numpy array
    npPre = np.array([ord(c) for c in prefix])

    # indicies for sliding window
    winIdx = np.arange(audio.size - len(prefix) + 1)[:, None] + np.arange(len(prefix))

    # find where the window matches
    tc_idx = (audio[winIdx] == npPre).all(1).nonzero()[0]
    
    #initialize empty arrays
    times = []
    samples = []

    for n,idx in enumerate(tc_idx):
        #audio index of string
        idx_str=idx
        #initialize time str
        tc_str = ''
        #loop till end of string
        while audio[idx_str] != 0:
            #get char from string
            tc_str += chr(audio[idx_str])
            #next index
            idx_str +=1
        
        time = datetime.datetime.strptime(tc_str, soft_time_fmt)
        # append time and index to array
        times.append(time)
        samples.append(idx)

    return np.array(times), np.array(samples)



import glob
import os
import re
import shutil
import stat
import subprocess
import tempfile
import warnings
from datetime import datetime


class log_search:

    """
    Class to parse and search MCV QoE log files

    The log_search class can read a single log file or a folder with multiple
    log files, addendum and group files.

    Attributes
    ----------
    fixedFields : list of strings
        The fixedFields static property is a list of fields that addendum files
        are not allowed to change
    found : set
        list of indicies matching the last search
    log : list of dicts
        the log attribute holds all of the parsed info from log files
    searchPath

    updateMode : {'Replace','AND','OR','XOR'}
        dictates how found is updated
    stringSearchMode : {'AND','OR','XOR'}
        dictates how string searches are performed
    foundCleared

    fieldNames : set
        a set of all the fieldnames that exist across all log entries
    """

    fixedFields = [
        "error",
        "complete",
        "operation",
        "GitHash",
        "logFile",
        "amendedBy",
        "date",
        "Arguments",
        "filename",
        "InputFile",
        "OutputFile",
    ]

    # Not hashable
    __hash__ = None

    def __init__(self, fname, addendumName=[], LogParseAction="Warn", groupName=[]):

        self.found = set()
        self.log = []
        self.searchPath = ""
        self.updateMode = "Replace"
        self.stringSearchMode = "OR"
        self.foundCleared = True
        self.fieldNames = set()
        self.groups = set()
        self.group_files = dict()

        if LogParseAction == "Warn":
            msgFcn = lambda m: warnings.warn(RuntimeWarning(m), stacklevel=3)
        elif LogParseAction == "Error":

            def errMsg(e):
                raise RuntimeError(e)

            msgFcn = errMsg
        elif LogParseAction == "Ignore":

            def noMsg(m):
                pass

            msgFcn = noMsg
        else:
            raise ValueError(f"invalid value '{LogParseAction}' for LogParseAction")

        if os.path.isdir(fname):
            # set searchpath to folder
            self.searchPath = fname

            # list log files in folder
            filenames = glob.glob(os.path.join(fname, "*.log"))

            # create full paths
            filenames = [os.path.join(fname, n) for n in filenames]

            # look for adendum files
            adnames = glob.glob(os.path.join(fname, "*.ad-log"))

            # check if we found any files
            if adnames:
                adnames[:] = [os.path.join(fname, n) for n in adnames]

            # look for group files
            groupnames = glob.glob(os.path.join(fname, "*.gr-log"))

            # check if we found any files
            if groupnames:
                groupnames[:] = [os.path.join(fname, n) for n in groupnames]

        else:

            (self.searchPath, _) = os.path.split(fname)

            # filenames is fname
            filenames = (fname,)

            if addendumName:
                adnames = (addendumName,)
            else:
                adnames = ()

            if groupName:
                groupnames = (groupName,)
            else:
                groupnames = ()

        # initialize idx
        idx = -1
        # -------------------------[Parse log files]----------------------------
        for fn in filenames:
            with open(fn, "r") as f:

                # create filename without path
                (_, short_name) = os.path.split(fn)

                # init status
                status = "searching"

                for lc, line in enumerate(f):

                    # strip newlines from line
                    line = line.strip("\r\n")

                    if line.startswith(">>"):
                        if not status == "searching":
                            msgFcn(
                                f"Start of packet found at line {lc} of {short_name} while in {status} mode"
                            )

                        # start of entry found, now we will parse the preamble
                        status = "preamble-st"

                    if status == "searching":
                        pass
                    elif status == "preamble-st":

                        # advance index
                        idx += 1

                        # add new dictionary to list
                        self.log.append({})

                        # split string into date and operation
                        parts = line.strip().split(" at ")

                        # set date
                        self.log[idx]["date"] = datetime.strptime(parts[1], "%d-%b-%Y %H:%M:%S")

                        # operation is the first bit
                        op = parts[0]

                        # remove >>'s from the beginning
                        op = op[2:]

                        # remove trailing ' started'
                        if op.endswith(" started"):
                            op = op[: -len(" started")]

                        # set operation
                        self.log[idx]["operation"] = op

                        # flag entry as incomplete
                        self.log[idx]["complete"] = False

                        # initialize error flag
                        self.log[idx]["error"] = False

                        # set source file
                        self.log[idx]["logFile"] = short_name

                        # dummy field for amendments
                        self.log[idx]["amendedBy"] = ""

                        # initialize groups
                        self.log[idx]["groups"] = set()

                        # set status to preamble
                        status = "preamble"
                    elif status == "preamble":

                        # check that the first character is not tab
                        if line and (not line[0] == "\t"):
                            # check for three equal signs
                            if not line.startswith("==="):
                                msgFcn(
                                    f"Unknown sequence found in preamble at line {lc} of file {short_name} : {repr(line)}"
                                )
                                # drop back to search mode
                                status = "searching"
                            elif line.startswith("===End"):
                                # end of entry, go into search mode
                                status = "searching"
                                # mark entry as complete
                                self.log[idx]["complete"] = True
                            elif line == "===Pre-Test Notes===":
                                status = "pre-notes"
                                # create empty pre test notes field
                                self.log[idx]["pre_notes"] = ""
                            elif line == "===Post-Test Notes===":
                                status = "post-notes"
                                # create empty post test notes field
                                self.log[idx]["post_notes"] = ""
                            else:
                                msgFcn(
                                    f"Unknown separator found in preamble at line {lc} of file {short_name} : {repr(line)}"
                                )
                                # drop back to search mode
                                status = "searching"
                        elif not line:
                            msgFcn(f"Empty line in preamble at line {lc} of file {short_name}")
                        else:
                            # split line on colon
                            lp = line.split(":")
                            # the MATLAB version uses genvarname but we just strip whitespace cause dict doesn't care
                            name = lp[0].strip()
                            # reconstruct the rest of line
                            arg = ":".join(lp[1:])

                            # check if key exists in dictionary
                            if name in self.log[idx].keys():
                                msgFcn(
                                    f"Duplicate field {name} found at line {lc} of file {short_name}"
                                )
                            else:
                                self.log[idx][name] = arg
                                # check for arguments
                                if name == "Arguments":
                                    self.log[idx]["_Arguments"] = self._argParse(arg)

                    elif status == "pre-notes":
                        # check that the first character is not tab
                        if line and line[0] == "\t":
                            # set sep based on if we have pre_notes
                            sep = "\n" if (self.log[idx]["pre_notes"]) else ""

                            # add in line, skip leading tab
                            self.log[idx]["pre_notes"] += sep + line.strip()
                        else:
                            # check for three equal signs
                            if not line.startswith("==="):
                                msgFcn(
                                    f"Unknown sequence found at line {lc} of file {short_name} : {repr(line)}"
                                )
                                # drop back to search mode
                                status = "searching"
                            elif line.startswith("===End"):
                                # end of entry, go into search mode
                                status = "searching"
                                # mark entry as complete
                                self.log[idx]["complete"] = True
                            else:
                                if line == "===Post-Test Notes===":
                                    status = "post-notes"
                                    # create empty post test notes field
                                    self.log[idx]["post_notes"] = ""
                                elif line == "===Test-Error Notes===":
                                    status = "post-notes"
                                    # create empty test error notes field
                                    self.log[idx]["error_notes"] = ""
                                    self.log[idx]["error"] = True
                                else:
                                    msgFcn(
                                        "Unknown separator found at line {lc} of file {short_name} : {repr(line)}"
                                    )
                                    # drop back to search mode
                                    status = "searching"

                    elif status == "post-notes":
                        # check that the first character is a tab
                        if line and line[0] == "\t":
                            field = "post_notes" if (not self.log[idx]["error"]) else "error_notes"

                            # set sep based on if we already have notes
                            sep = "" if (self.log[idx][field]) else "\n"

                            # add in line, skip leading tab
                            self.log[idx][field] += sep + line.strip()
                        else:
                            # check for three equal signs
                            if not line.startswith("==="):
                                msgFcn(
                                    f"Unknown sequence found at line {lc} of file {short_name} : {repr(line)}"
                                )
                                # drop back to search mode
                                status = "searching"
                            elif line.startswith("===End"):
                                # end of entry, go into search mode
                                status = "searching"
                                # mark entry as complete
                                self.log[idx]["complete"] = True
                            else:
                                msgFcn(
                                    f"Unknown separator found at line {lc} of file {short_name} : {repr(line)}"
                                )
                                # drop back to search mode
                                status = "searching"
        # -------------------------[Parse addendum files]-----------------------
        for fn in adnames:
            with open(fn, "r") as f:

                # create filename without path
                (_, short_name) = os.path.split(fn)

                # init status
                status = "searching"

                for lc, line in enumerate(f):
                    # always check for start of entry
                    if line.startswith(">>"):
                        # check if we were in search mode
                        if not status == "searching":
                            # give error when start found out of sequence
                            raise ValueError(
                                f"Start of addendum found at line {lc} of file {short_name} while in {status} mode"
                            )

                        # start of entry found, now we will parse the preamble
                        status = "preamble-st"

                    if status == "searching":
                        pass
                    elif status == "preamble-st":

                        # split string into date and operation
                        parts = line.strip().split(" at ")

                        # set date
                        date = datetime.strptime(parts[1], "%d-%b-%Y %H:%M:%S")

                        # operation is the first bit
                        op = parts[0]

                        # remove >>'s from the beginning
                        op = op[2:]

                        # remove trailing ' started'
                        if op.endswith(" started"):
                            op = op[: -len(" started")]

                        # get index from _logMatch
                        idx = self._logMatch({"date": date, "operation": op})

                        if not idx:
                            raise ValueError(
                                f"no matching entry found for '{line.strip()}' from file {short_name}"
                            )
                        elif len(idx) > 1:
                            raise ValueError(
                                f"multiple matching entries found for '{line.strip()}' from file {short_name}"
                            )

                        # get index from set
                        idx = idx.pop()

                        # check if this log entry has been amended
                        if self.log[idx]["amendedBy"]:
                            # indicate which file amended this entry
                            self.log[idx]["amendedBy"] = short_name
                        else:
                            ValueError(f"log entry already amended at line {lc} of '{short_name}'")

                        # set status to preamble
                        status = "preamble"

                    elif status == "preamble":

                        # check that the first character is not tab
                        if line and not line[0] == "\t":
                            # check for end marker
                            if not line.startswith("<<"):
                                raise ValueError(
                                    f"Unknown sequence found in entry at line {lc} of file {short_name} : {line}"
                                )
                            else:
                                # end of entry, go into search mode
                                status = "searching"
                        else:
                            # split line on colon
                            lp = line.split(":")

                            # the MATLAB version uses genvarname but we just strip whitespace cause dict doesn't care
                            name = lp[0].strip()
                            # reconstruct the rest of line
                            arg = ":".join(lp[1:])

                            # check if field is amendable
                            if name in self.fixedFields:
                                raise ValueError(
                                    f"At line {lc} of file {short_name} : field '{name}' is not amendable"
                                )

                            # check if field exists in structure
                            if name in self.log[idx].keys():
                                self.log[idx][name] = arg
                            else:
                                raise ValueError(
                                    f"Invalid field {repr(name)} at line {lc} of {short_name}"
                                )
        # -------------------------[Parse Group Files]-------------------------
        for fn in groupnames:
            with open(fn, "r") as f:

                # create filename without path
                (_, short_name) = os.path.split(fn)

                # Initialize dictionary for this file name
                self.group_files[short_name] = set()

                for lc, line in enumerate(f):
                    # remove whitespace
                    line = line.strip()

                    # check for blank lines
                    if not line:
                        continue

                    # check for comments
                    if line[0] == "#":
                        continue

                    parts = line.split(":")

                    groupNames = parts[0].strip().split(",")

                    members = ":".join(parts[1:]).split(",")

                    for ds in members:

                        # set date
                        date = datetime.strptime(ds.strip(), "%d-%b-%Y %H:%M:%S")

                        idx = self._logMatch({"date": date})

                        if not idx:
                            raise ValueError(
                                f"no matching entry found for '{line.strip()}' from file {short_name}"
                            )
                        elif len(idx) > 1:
                            raise ValueError(
                                f"multiple matching entries found for '{line.strip()}' from file {short_name}"
                            )

                        for groupName in groupNames:
                            # strip leading and trailing spaces
                            groupName = groupName.strip()
                            # contcatenate filename and groupname
                            full_group_name = "{}:{}".format(short_name, groupName)
                            self.groups.add(full_group_name)
                            self.group_files[short_name].add(full_group_name)

                            self.log[list(idx)[0]]["groups"].add(full_group_name)

        for l in self.log:
            self.fieldNames.update(l.keys())

    def _logMatch(self, match):
        """
        Internal function to find matching log entries
        """
        m = set()
        for n, x in enumerate(self.log):
            eq = [False] * len(match)
            for i, k in enumerate(match.keys()):
                if k == "date_before":
                    eq[i] = match[k] > x["date"]
                elif k == "date_after":
                    eq[i] = match[k] < x["date"]
                else:
                    try:
                        val = x[k]
                    except KeyError:
                        eq[i] = False
                        # done here, continue
                        continue
                    # check for strings and handle differently
                    if isinstance(val, str):
                        if isinstance(match[k], list):
                            str_eq = [(re.compile(s).search(val)) is not None for s in match[k]]
                            if self.stringSearchMode == "AND":
                                eq[i] = all(str_eq)
                            elif self.stringSearchMode == "OR":
                                eq[i] = any(str_eq)
                            elif self.stringSearchMode == "XOR":
                                # check if exactly one string matched
                                eq[i] = 1 == str_eq.count(True)
                        else:
                            eq[i] = re.compile(match[k]).search(val) is not None
                    elif isinstance(val, set):
                        if isinstance(match[k], set):
                            eq[i] = match[k].issubset(val)
                        else:
                            eq[i] = match[k] in val
                    else:
                        # fall back to equals comparison
                        eq[i] = val == match[k]
            if all(eq):
                m.add(n)
        return m

    def _foundUpdate(self, idx):
        """
        internal function to update the found attribute based on the update mode
        """
        # make idx a set
        idx = set(idx)
        if self.foundCleared:
            self.found = idx
        else:
            if self.updateMode == "Replace":
                self.found = idx
            elif self.updateMode == "AND":
                self.found &= idx
            elif self.updateMode == "OR":
                self.found |= idx
            elif self.updateMode == "XOR":
                self.found ^= idx
            else:
                raise ValueError(f"Unknown updateMode '{self.updateMode}'")
        # clear cleared
        self.foundCleared = False

    def Qsearch(self, search_field, search_term):
        """
        Quick search : search for a match in a single property

        Parameters
        ----------
        search_field : string
            The name of the field to search
        search_term
            the value that the search field values will be compared to

        Returns
        -------
        The set of matching indicies in self.log
        """

        # find matching entries
        idx = self._logMatch({search_field: search_term})

        # update found array
        self._foundUpdate(idx)

        return idx

    def MfSearch(self, search):
        """
        Multi field search : search for multiple matching fields

        Parameters
        ----------
        search : dict
            a dictionary where the keys are the fields to search and the values are
            the things to search for

        Returns
        -------
        The set of matching indicies in self.log
        """

        # search must be a dictionary
        if not isinstance(search, dict):
            raise ValueError("the search argument must be a dictionary")

        # search for matching log entries
        idx = self._logMatch(search)

        # update found array
        self._foundUpdate(idx)

        return idx

    def clear(self):
        """
        clear the found set
        """
        # clear found
        self.found = set()
        self.foundCleared = True

    def datafilenames(self, ftype="csv",ignore_incomplete=False):
        """
        find data files matching a log entry

        Parameters
        ----------
        ftype : {'mat','csv','bad_csv','wav','sm_mat'}
            what type of files to look for
        """

        #name of test opperations
        test_opps=('Test','Intelligibility','PSuD','Access','M2E')

        types = re.compile(
            r"\.?(?P<csv>csv)|(?P<mat>mat)|(?P<wav>wav)|(?P<sm_mat>sm(?:all)?_mat)|(?P<bad_csv>bad_csv)",
            re.IGNORECASE,
        )

        m = types.match(ftype)

        if not m:
            raise ValueError(f"Unknown search type '{ftype}'")

        if m.group("mat"):
            # TODO: Should we delete this?
            tstFiles = {
                "ext": ".mat",
                "path": "data_matfiles",
                "singular": True,
                "exclude": "",
            }
        elif m.group("csv"):
            tstFiles = {
                "ext": ".csv",
                "path": os.path.join("data", "csv"),
                "singular": False,
                "exclude": "_BAD.csv",
            }
        elif m.group("bad_csv"):
            # TODO: where is this now?
            tstFiles = {
                "ext": "_BAD.csv",
                "path": os.path.join("post-processed data", "csv"),
                "singular": False,
                "exclude": None,
            }
        elif m.group("wav"):
            tstFiles = {
                "ext": "",
                "path": os.path.join("data", "wav"),
                "singular": True,
                "exclude": None,
            }
        elif m.group("sm_mat"):
            # TODO: Should we delete this?
            tstFiles = {
                "ext": ".mat",
                "path": os.path.join("post-processed data", "mat"),
                "singular": True,
                "exclude": None,
            }
        else:
            raise RuntimeError(f"'{ftype}' is an invalid file type")

        # filenames
        fn = []
        # file indexes
        fi = []
        for idx in self.found:
            if self.log[idx]["operation"] in test_opps:
                prefix = ["Rcapture_", "capture_"]
                folder = [tstFiles["path"]] * len(prefix)
                ext = tstFiles["ext"]
                singular = tstFiles["singular"]
                exclude = tstFiles["exclude"]
            elif self.log[idx]["operation"] == "Training":
                # TODO: Can we delete training? I think so...
                prefix = ["Training_"] * 2
                folder = ["training", "data"]
                ext = ".mat"
                singular = True
                exclude = None
            elif self.log[idx]["operation"] == "Tx Two Loc Test":
                # TODO: Does this need an update?
                prefix = ["Tx_capture", "capture"]
                folder = ["tx-data"] * len(prefix)
                ext = ".mat"
                singular = True
                exclude = None
            elif self.log[idx]["operation"] == "Rx Two Loc Test":
                # TODO: Does this need an update?
                prefix = ["Rx_capture", "capture"]
                folder = ["rx-data"] * len(prefix)
                ext = ".mat"
                singular = True
                exclude = None
            elif self.log[idx]["operation"].startswith("Copy"):
                fn.append(":None")
                fi.append(idx)
                continue
            else:
                raise ValueError(f"Unknown operation '{self.log[idx]['operation']}'")

            if (not ignore_incomplete) and (not self.log[idx]["complete"]):
                fn.append(":Incomplete")
                fi.append(idx)
                continue

            if self.log[idx]["error"]:
                fn.append(":Error")
                fi.append(idx)
                continue

            # get date string in file format
            date_str = self.log[idx]["date"].strftime("%d-%b-%Y_%H-%M-%S")

            for f, p in zip(folder, prefix):

                foldPath = os.path.join(self.searchPath, f)

                filenames = glob.glob(os.path.join(foldPath, p + "*" + ext))

                match = [
                    f for f in filenames if date_str in f and ((not exclude) or exclude not in f)
                ]

                if not singular and len(match) >= 1:
                    fn += match
                    fi.append([idx] * len(match))
                    break
                elif len(match) > 1:
                    print(f"More than one file found matching '{date_str}' in '{f}")
                    fn.append("Multiple")
                    fi.append(idx)
                elif len(match) == 1:
                    fn.append(match[0])
                    fi.append(idx)
                    break
            else:
                fn.append(None)
                fi.append(idx)
                warnings.warn(
                    RuntimeWarning(f"No matching files for '{date_str}' in '{foldPath}'"),
                    stacklevel=2,
                )

        return (fn, fi)

    def findFiles(self, locpath, ftype="csv"):
        """Get filenames from current log search object.

        Similar to datafilenames, but copies any found files that are in the
        network search path but missing from the local path into the local
        path.

        Parameters
        ----------
        locpath : string
            Local path for data storage. .
        ftype : string, optional
            File type for data to search for/copy. The default is 'csv'.

        Returns
        -------
        filenames : list
            List of filenames that match log search conditions.Filenames are
            guaranteed to exist locally.

        """

        network_path = self.searchPath

        #       self.searchPath = network_path
        net_names, net_ix = self.datafilenames(ftype)

        # Identify all sessions marked error on network
        net_errSessions = [name == ":Error" for name in net_names]

        # Identify all sessions marked Incomplete on network
        net_incSessions = [name == ":Incomplete" for name in net_names]

        # Identify all sessoins that could not be identified on netowkr
        net_notFound = [name == None for name in net_names]

        if any(net_notFound):
            warnings.warn(
                RuntimeWarning(f"'{sum(net_notFound)}' files not found on network"),
                stacklevel=2,
            )

        net_tossIx = [
            net_errSessions[i] or net_incSessions[i] or net_notFound[i]
            for i in range(0, len(net_names))
        ]
        # Toss unwanted sessions from net_names
        net_names_clean = [net_names[i] for i in range(0, len(net_names)) if (not (net_tossIx[i]))]

        if os.path.exists(os.path.join(locpath, "post-processed data", ftype)):
            # Check that path to data even exists

            # Switch to searching localpath
            self.searchPath = locpath
            loc_names, loc_ix = self.datafilenames(ftype)

            # Identify all sessions marked error/incomplete locally
            loc_errSessions = [name == ":Error" for name in loc_names]
            loc_incSessions = [name == ":Incomplete" for name in loc_names]
            loc_NoneSessions = [name == None for name in loc_names]
            loc_tossIx = [
                loc_NoneSessions[i] or loc_errSessions[i] or loc_incSessions[i]
                for i in range(0, len(loc_names))
            ]
            loc_names_clean = [
                loc_names[i] for i in range(0, len(loc_names)) if (not (loc_tossIx[i]))
            ]
        else:
            # Path to data does not exist at all locally
            # Set local names to be empty
            loc_names_clean = []

        #       # Grab just the basenames of each list
        loc_base = [os.path.basename(x) for x in loc_names_clean]
        net_base = [os.path.basename(x) for x in net_names_clean]

        # Transform into sets
        loc_set = set(loc_base)
        net_set = set(net_base)
        # Take set union of local and network names
        all_names = loc_set.union(net_set)
        # Identify names in network that are missing from local set
        non_localNames = all_names.difference(loc_set)

        for fname in non_localNames:
            _, fbase = os.path.split(fname)
            if ftype == "csv":
                subdir = os.path.join("post-processed data", "csv")
            elif ftype == "sm_mat":
                subdir = os.path.join("post-processed data", "mat")
            elif ftype == "wav":
                subdir = os.path.join("post-processed data", "wav")
            elif ftype == "mat":
                subdir = "data"
            else:
                raise RuntimeError(f"Unsupported ftype: '{ftype}' ")

            lpath = os.path.join(locpath, subdir)
            # Make directories if don't exist locally
            if not (os.path.exists(lpath)):
                os.makedirs(lpath)

            localpath = os.path.join(lpath, fname)
            netpath = os.path.join(network_path, subdir, fname)
            print(f"Copying from:\n -- {netpath}", flush=True)
            print(f"Copying to:\n -- {localpath}", flush=True)
            shutil.copy2(netpath, localpath)

        filenames, _ = self.datafilenames(ftype)

        f_errSessions = [name == ":Error" for name in filenames]
        f_incSessions = [name == ":Incomplete" for name in filenames]
        f_tossIx = [f_errSessions[i] or f_incSessions[i] for i in range(0, len(filenames))]
        filenames = [filenames[i] for i in range(0, len(filenames)) if (not (f_tossIx[i]))]

        if filenames == []:
            raise RuntimeError("Could not find any files meeting search criteria")
        self.searchPath = network_path
        return filenames

    def isAncestor(self, rev, repo_path, git_path=None):
        """
        Search for testss run with an ancestor of the given rev

        This searches the git hash field and can be used to find tests that have
        been run with code that is more recent than `rev`. This requires the
        repository that the tests were run using.

        Parameters
        ----------
        rev : string
            string that can be resolved, by git, to a commit. Posible values include
            branch names, tag names, and partial commit hashes
        repo_path : string
            this can either be a local path to the repository or a git URL. If
            it is a git URL the repository will be cloned and used for the search.
        git_path : string
            the path to the git executable

        Returns
        -------
        The set of matching indices in self.log
        """

        if git_path is None:
            git_path = "git"

        # dummy name in case it is not used
        tmpdir = None
        try:

            if isGitURL(repo_path):
                # save URL
                repo_url = repo_path
                # generate temp dir
                tmpdir = tempfile.TemporaryDirectory()
                # print message to inform the user
                print(f"Cloning : {repo_path}")
                # set repo path to temp dir
                repo_path = tmpdir.name
                # clone repo
                p = subprocess.run([git_path, "clone", repo_url, repo_path], capture_output=True)
                # check return code
                if p.returncode:
                    # TODO: error
                    raise RuntimeError(p.stderr)

            # get the full has of the commit described by rev
            p = subprocess.run(
                [git_path, "-C", repo_path, "rev-parse", "--verify", rev],
                capture_output=True,
            )
            # convert to string and trim whitespace
            rev_p = p.stdout.decode("utf-8").strip()
            # check for success
            if p.returncode:
                raise RuntimeError(f"Failed to parse rev : {rev_p}")

            match = set()

            hashCache = {}

            for k, l in enumerate(self.log):
                hash = l["Git Hash"].strip()

                if not hash:
                    # skip this one
                    continue

                # dump dty flag
                hash = hash.split()[0]

                if hash not in hashCache.keys():

                    # check that hash is valid
                    p = subprocess.run(
                        [git_path, "-C", repo_path, "cat-file", "-t", hash],
                        capture_output=True,
                    )

                    if p.returncode:
                        warnings.warn(
                            RuntimeWarning(
                                f"Could not get hash for {hash} {p.stderr.decode('utf-8').strip()}"
                            ),
                            stacklevel=2,
                        )
                        # store result in hash cache
                        hashCache[hash] = False
                        # skip this one
                        continue

                    p = subprocess.run(
                        [
                            git_path,
                            "-C",
                            repo_path,
                            "show-branch",
                            "--merge-base",
                            rev_p,
                            hash,
                        ],
                        capture_output=True,
                    )
                    # remove spaces from base
                    base = p.stdout.decode("utf-8").strip()
                    # check for errors
                    if p.returncode:
                        warnings.warn(
                            RuntimeWarning(
                                f"Could not get the status of log entry {k} git returned : {base}"
                            ),
                            stacklevel=2,
                        )

                    hashCache[hash] = base == rev_p

                if hashCache[hash]:
                    match.add(k)
        finally:
            if tmpdir:
                # must delete directory manually because of a bug in TemporaryDirectory
                # see https://bugs.python.org/issue26660
                # .git has read only files
                shutil.rmtree(os.path.join(tmpdir.name, ".git"), onerror=del_rw)
                # cleanup dir
                tmpdir.cleanup()

        self._foundUpdate(match)

        return match

    def argSearch(self, name, value):
        """
        search the arguments field of log entries

        Searches the arguments field for arguments that match the given name, value
        pair. Note : this has not been well tested with python, enjoy!

        Parameters
        ----------
        name : string
            the parameter name to search for
        value : string
            the parameter value to search for
        """

        def listCmp(l1, l2):

            m = [False] * len(l2)

            for a in l1:
                res = [valCmp(a, b) for b in l2]

                # make sure that there was a match
                if not any(res):
                    return False

                # OR lists together
                m = [a | b for a, b in zip(m, res)]

            return all(m)

        def valCmp(arg, val):
            if isinstance(arg, list) and isinstance(val, list):
                return listCmp(arg, val)
            elif isinstance(arg, list) and not isinstance(val, list):
                return listCmp(arg, [val])
            elif (not isinstance(arg, list)) and isinstance(val, list):
                return listCmp([arg], val)
            elif isinstance(arg, str):
                m = re.search(val, arg)
                return not not m
            else:
                return arg == val

        match = set()
        for i, l in enumerate(self.log):
            try:
                if valCmp(l["_Arguments"][name], value):
                    match.add(i)
            except KeyError:
                # argument not found, not a match
                pass

        self._foundUpdate(match)

        return match

    def _argParse(self, args):
        """
        internal function for parsing the arguments field

        This has not been well tested for python generated log files so, it may not
        work so well.
        """

        def str_or_float(val):
            if not val:
                return None
            m = re.match(r"(?P<str>'(?P<s>[^']*)')|(?P<true>true)|(?P<false>false)", val)
            if m:
                if m.group("str"):
                    return m.group("s")
                elif m.group("true"):
                    return True
                elif m.group("false"):
                    return False
                else:
                    raise RuntimeError("Internal Error")
            else:
                try:
                    return float(val)
                except ValueError:
                    warnings.warn(RuntimeWarning(f"Could not convert '{val}'"), stacklevel=2)
                    return val

        match_args = re.finditer(
            r"'(?P<name>[^']*)',(?P<value>(?P<cell_m>\{(?P<cell>[^}]*)\})|(?P<arr_m>\[(?P<arr>[^]]*)\])|(?:[^{[][^,]*))",
            args,
        )

        # dictionary for args
        arg_d = {}

        for m in match_args:
            # check for cell array
            if m.group("cell_m"):
                if m.group("cell"):
                    arg_d[m.group("name")] = [
                        str_or_float(v) for v in re.split(";|,", m.group("cell"))
                    ]
                else:
                    arg_d[m.group("name")] = []
            # check for a array
            elif m.group("arr_m"):
                if m.group("arr"):
                    arg_d[m.group("name")] = [
                        str_or_float(v) for v in re.split(";|,", m.group("arr"))
                    ]
                else:
                    arg_d[m.group("name")] = []
            else:
                arg_d[m.group("name")] = str_or_float(m.group("value"))

        return arg_d

    def argQuery(self, argName):
        """
        I really don't remember what this is for
        """

        f_log = self.flog

        res = [None] * len(f_log)
        for i, entry in enumerate(f_log):
            res[i] = entry["_Arguments"][argName]

        return res

    def __len__(self):
        # return number of log entries
        return len(self.log)

    @property
    def flog(self):
        return [self.log[i] for i in self.found]


# workaround for deleting read only files
# code from : https://bugs.python.org/issue19643#msg208662
def del_rw(action, name, exc):
    """
    workaround for deleting read only files with shutil.rmtree
    """
    os.chmod(name, stat.S_IWRITE)
    os.remove(name)


def isGitURL(str):
    """
    detect if a url could be to a git repository
    """
    if str.startswith("git@"):
        return True
    elif str.startswith("https://"):
        return True
    elif str.startswith("http://"):
        return True
    else:
        return False
